#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <dirent.h>
#include <linux/input.h>
#include <linux/uinput.h>
#include <linux/kd.h>
#include <limits.h>
#include <fcntl.h>
#include <errno.h>

#define INPUTDIR "/dev/input"
#define TABLESIZE 25359
#define CAPITAL(ch) \
        if(shift) break; \
        if(ev.value) { tmp[i++]=ch; \
        if(i>5) { write(1,"\a",1); i=0; } \
        if(grave) { tmp[i]='\0'; commit(tmp); grave=0; } } continue;
#define DIGIT(ch) \
        if(i==0) break; \
        if(ev.value) { tmp[i++]=ch; tmp[i]='\0'; commit(tmp); } continue;
#define BRACE(ch) \
        if(shift) { if(ev.value) { tmp[i++]=ch; continue; } else break; } break;
char bsm[TABLESIZE][2][6],in[6],out[6],tmp[6];
int input,output,tty;
int i;

int commit(char *word)
{
  char *big5;
  tty=open("/dev/tty0",O_RDONLY);
  for(i=0; i<TABLESIZE; i++)
  {
    if(strcmp(bsm[i][0],word)==0)
    {
      for(big5=bsm[i][1]; *big5; big5++)
        ioctl(tty,TIOCSTI,big5);
      break;
    }
  }
  if(i==TABLESIZE) write(1,"\a",1);
  close(tty);
  i=0;
  return 0;
}

int main()
{
  printf("Boshamy for linux console by Coscell\n");
//---- Load table ----------------------
  FILE *file=fopen("bsm.txt","r");
  if(!file) { printf("\absm.txt not found!\n"); exit(-1); }

  for(i=0; i<TABLESIZE; i++)
  {
    fscanf(file,"%s%s",in,out);
    strcpy(bsm[i][0],in);
    strcpy(bsm[i][1],out);
  }
  fclose(file);
//---- I/O Initialization --------------
  int error=0;
  struct dirent *entry;
  DIR *dir=opendir(INPUTDIR);
  if(!dir) { printf("\a%s: No such directory!\n",INPUTDIR); exit(-1); }

  while(entry=readdir(dir))
  {
    char buf[strlen(INPUTDIR)+1+strlen(entry->d_name)+1];
    unsigned char keys[(KEY_MAX+1)/CHAR_BIT];
    snprintf(buf,sizeof(buf),"%s/%s",INPUTDIR,entry->d_name);
    input=open(buf,O_RDONLY);
    if(input==-1) { printf("\aOpen %s failed! Try next ... \n",buf); error=1; continue; }
    ioctl(input,EVIOCGBIT(EV_KEY,KEY_MAX),&keys);
    if(!(keys[(KEY_A)/CHAR_BIT]&(1<<((KEY_A)%CHAR_BIT))))
    {
      close(input); input=-1; continue; \
    }
    ioctl(input,EVIOCGRAB,1);
    if(error) printf("Successful!\n");
    break;
  }
  if(!entry) { printf("\n\aNo input devices!\n"); exit(-1); }

  struct uinput_user_dev setup={};
  int key;
  while(1)
  {
    output=open("/dev/uinput",O_RDWR);
    if(output>=0) break;
    if(errno!=ENOENT) { perror("open"); exit(-1); }
    sleep(1);
  }
  strncpy(setup.name,"bsm",sizeof(setup.name));
  if(write(output,&setup,sizeof(setup))==-1 ||
     ioctl(output,UI_SET_EVBIT,EV_KEY)==-1 ||
     ioctl(output,UI_SET_EVBIT,EV_REP)==-1)
  {
    printf("\aInput devices error!\n");
    exit(-1);
  }

  for(key=0; key<128; key++) ioctl(output,UI_SET_KEYBIT,key);
  if(ioctl(output,UI_DEV_CREATE)==-1)
  {
    printf("\aInput devices failed!\n");
    exit(-1);
  }
//---- Boshamy -------------------------
  system("beep -l100 -f 880 -n -l100 -f 660");
  struct input_event ev;
  char shift=0,grave=0,mode=0,led;
  i=0;

  pid_t pid=fork();
  if(pid) exit(1);

  while(1)
  {
    read(input,&ev,sizeof(ev));
    if((ev.code==KEY_LEFTMETA || ev.code==KEY_RIGHTMETA) && ev.value)
    {
      mode=(!mode);
      system(mode?"beep -l100 -f 660 -n -l100 -f 880":"beep -l100 -f 880 -n -l100 -f 660");
      if(!mode) i=0;
    }
    if(ev.code==KEY_CAPSLOCK && ev.value)
    {
      tty=open("/dev/tty0", O_RDWR);
      ioctl(tty,KDGETLED,&led);
      close(tty);
      system(!(led&LED_CAP)?"beep -l100 -f 660 -n -l100 -f 880":"beep -l100 -f 880 -n -l100 -f 660");
      close(tty);
    }

    if(mode)
    {
      switch(ev.code)
      {
        case KEY_LEFTALT:
        case KEY_RIGHTALT:
        case KEY_LEFTCTRL:
        case KEY_RIGHTCTRL:
        case KEY_LEFTSHIFT:
        case KEY_RIGHTSHIFT: shift=ev.value?1:0; break;
        case KEY_A: CAPITAL('a');
        case KEY_B: CAPITAL('b');
        case KEY_C: CAPITAL('c');
        case KEY_D: CAPITAL('d');
        case KEY_E: CAPITAL('e');
        case KEY_F: CAPITAL('f');
        case KEY_G: CAPITAL('g');
        case KEY_H: CAPITAL('h');
        case KEY_I: CAPITAL('i');
        case KEY_J: CAPITAL('j');
        case KEY_K: CAPITAL('k');
        case KEY_L: CAPITAL('l');
        case KEY_M: CAPITAL('m');
        case KEY_N: CAPITAL('n');
        case KEY_O: CAPITAL('o');
        case KEY_P: CAPITAL('p');
        case KEY_Q: CAPITAL('q');
        case KEY_R: CAPITAL('r');
        case KEY_S: CAPITAL('s');
        case KEY_T: CAPITAL('t');
        case KEY_U: CAPITAL('u');
        case KEY_V: CAPITAL('v');
        case KEY_W: CAPITAL('w');
        case KEY_X: CAPITAL('x');
        case KEY_Y: CAPITAL('y');
        case KEY_Z: CAPITAL('z');
        case KEY_COMMA: CAPITAL(',');
        case KEY_DOT: CAPITAL('.');
        case KEY_GRAVE: if(ev.value) { grave=1; tmp[i++]='`'; } continue;
        case KEY_1: DIGIT('1');
        case KEY_2: DIGIT('2');
        case KEY_3: DIGIT('3');
        case KEY_4: DIGIT('4');
        case KEY_LEFTBRACE: BRACE('{');
        case KEY_RIGHTBRACE: BRACE('}');
        case KEY_APOSTROPHE: BRACE('"');
        case KEY_SPACE: if(i==0) break;
             if(ev.value) { tmp[i]='\0'; commit(tmp); if(grave) grave=0; } continue;
        case KEY_BACKSPACE: if(i) { if(ev.value) i--; continue; }
      }
    }
    write(output,&ev,sizeof(ev));
  }
  system("beep -l100 -f 880 -n -l100 -f 660");
  close(input);
  close(output);
  wait(NULL);
  return 0;
}
